#!/usr/bin/python
# encoding: utf-8
#
# Copyright (C) 2010 Marco A L Barbosa <malbarbo@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.


# Este programa ofereçe diversas funções relacionadas a linguagens
# regulares, como simulação de AFD e AFN.  O código deste programa foi
# escrito de maneira a deixar claro a relação com a teoria (definições
# formais e algoritmos) e não se preocupa com otimizações.  A entrada de
# dados foi escrita de maneira simples, e não faz muitas checagens de
# erro.


# Constantes

EPSLON = ''
MARCA_INICIAL = '>'
MARCA_FINAL = '*'
ACEITA = 'accept'
REJEITA = 'reject'
SEP_COLUNA = '|'
SEP_ESTADO = ' '


# AFD - Autômato finito determinístico

class AFD:
    def __init__(self, estados, alfabeto, transicoes, inicial, finais):
        self.estados = estados
        self.alfabeto = alfabeto
        self.transicoes = transicoes
        self.inicial = inicial
        self.finais = finais
    
    def aceita(automato, cadeia):
        ativo = automato.inicial
        for simbolo in cadeia:
           ativo = automato.transicoes[ativo][simbolo]
        return ativo in automato.finais
    
def afd_ler_da_tabela(tabela):
    estados, inicial, finais = ler_estados(tabela)
    alfabeto = ler_alfabeto(tabela)
    transicoes = afd_ler_transicoes(tabela, estados, alfabeto)
    return AFD(estados, alfabeto, transicoes, inicial, finais)

def afd_ler_transicoes(tabela, estados, alfabeto):
    transicoes = {}
    for i in range(0, len(estados)):
        transicoes[estados[i]] = {}
        for j in range(0, len(alfabeto)):
            transicoes[estados[i]][alfabeto[j]] = tabela[i+1][j+1]
    return transicoes

def afd_ler_do_arquivo(nome_arquivo):
    linhas = ler_linhas(nome_arquivo)
    tabela = ler_tabela(linhas)
    automato = afd_ler_da_tabela(tabela)
    return automato

# AFN - Autômato finito não-determinístico

class AFN:
    def __init__(self, estados, alfabeto, transicoes, inicial, finais):
        self.estados = estados
        self.alfabeto = alfabeto
        self.transicoes = transicoes
        self.inicial = inicial
        self.finais = finais
        self.e_fechamento = {}
        self.calcular_todos_e_fechamento()

    def calcular_todos_e_fechamento(self):
        for estado in self.estados:
            self.calcula_e_fechamento(estado)
    
    def calcula_e_fechamento(self, estado):
        fechamento = [estado]
        i = 0
        while i < len(fechamento):
            q = fechamento[i]
            for p in self.transicoes[q][EPSLON]:
                if p not in fechamento:
                    fechamento.append(p)
            i = i + 1
        self.e_fechamento[estado] = set(fechamento)

    def aceita(automato, cadeia):
        ativos = automato.e_fechamento[automato.inicial]
        for simbolo in cadeia:
            novos_ativos = set()
            for estado in ativos:
                destinos = automato.transicoes[estado][simbolo]
                for destino in destinos:
                    novos_ativos = novos_ativos.union(automato.e_fechamento[destino])
            ativos = novos_ativos
        for estado in ativos:
            if estado in automato.finais:
                return True
        return False

def afn_ler_da_tabela(tabela):
    estados, inicial, finais = ler_estados(tabela)
    alfabeto = ler_alfabeto(tabela)
    transicoes = afn_ler_transicoes(tabela, estados, alfabeto)
    return AFN(estados, alfabeto, transicoes, inicial, finais)

def afn_ler_transicoes(tabela, estados, alfabeto):
    transicoes = {}
    for i in range(0, len(estados)):
        transicoes[estados[i]] = {}
        for j in range(0, len(alfabeto)):
            if tabela[i+1][j+1] == '':
                transicoes[estados[i]][alfabeto[j]] = []
            else:
                transicoes[estados[i]][alfabeto[j]] = separar(tabela[i+1][j+1], SEP_ESTADO)
    return transicoes

def afn_ler_do_arquivo(nome_arquivo):
    linhas = ler_linhas(nome_arquivo)
    tabela = ler_tabela(linhas)
    automato = afn_ler_da_tabela(tabela)
    return automato


# Funções comuns para AFD e AFN

def ler_estados(tabela):
    estados = []
    inicial = None
    finais = []
    for i in range(1, len(tabela)):
        estado = tabela[i][0].replace(MARCA_INICIAL, '').replace(MARCA_FINAL, '')
        if tabela[i][0].startswith(MARCA_INICIAL):
            inicial = estado
        if tabela[i][0].endswith(MARCA_FINAL):
            finais.append(estado)
        estados.append(estado)
    return estados, inicial, finais

def ler_alfabeto(tabela):
    return tabela[0][1:]

def ler_cadeias_de_teste(nome_arquivo):
    aceita = []
    rejeita = []
    for linha in ler_linhas(nome_arquivo):
        if linha.startswith(ACEITA):
            aceita.append(linha[6:].strip())
        elif linha.startswith(REJEITA):
            rejeita.append(linha[6:].strip())
    return aceita, rejeita

def testar_automato(automato, aceita, rejeita):
    for cadeia in aceita:
        if automato.aceita(cadeia):
            print 'OK   aceita ', cadeia
        else:
            print 'FAIL aceita ', cadeia
    for cadeia in rejeita:
        if automato.aceita(cadeia):
            print 'FAIL rejeita', cadeia
        else:
            print 'OK   rejeita', cadeia

def imprimir_automato(automato):
    tabela = []

    linha = ['']
    for simbolo in automato.alfabeto:
        linha.append(simbolo)
    tabela.append(linha)

    for estado in automato.estados:
        s = estado
        if estado == automato.inicial:
            s = MARCA_INICIAL + s
        else:
            s = ' ' + s
        if estado in automato.finais:
            s = s + MARCA_FINAL
        linha = [s]
        for simbolo in automato.alfabeto:
            destino = automato.transicoes[estado][simbolo]
            if type(destino) == list:
                destino = SEP_ESTADO.join(destino)
            linha.append(destino)
        tabela.append(linha)

    max_coluna = [0] * len(tabela[0])
    for i in range(0, len(tabela)):
        for j in range(0, len(tabela[i])):
            max_coluna[j] = max(max_coluna[j], len(tabela[i][j]))
    
    for i in range(0, len(tabela)):
        linha = []
        for j in range(0, len(tabela[i])):
            celula = tabela[i][j]
            celula = celula + ' ' * (max_coluna[j] - len(celula))
            linha.append(celula)
        print (' ' + SEP_COLUNA + ' ').join(linha)


# Funções para leitura de arquivos

def ler_tabela(linhas):
    tabela = []
    for linha in linhas:
        if linha == '':
            break        
        tabela.append(separar(linha, SEP_COLUNA))
    return tabela

def separar(linha, separador):
    return [v.strip() for v in linha.split(separador)]

def ler_linhas(nome_arquivo):
    arquivo = open(nome_arquivo)
    linhas = [linha.strip() for linha in arquivo.readlines()]
    arquivo.close()
    return linhas


# Comandos

def testar_afd(params):
    """Lê um afd a partir de um arquivo e testas a pertinência de cadeias."""
    automato = afd_ler_do_arquivo(params[0])
    aceita, rejeita = ler_cadeias_de_teste(params[0])
    imprimir_automato(automato)
    print
    testar_automato(automato, aceita, rejeita)

def testar_afn(params):
    """Lê um afn a partir de um arquivo e testas a pertinência de cadeias."""
    automato = afn_ler_do_arquivo(params[0])
    aceita, rejeita = ler_cadeias_de_teste(params[0])
    imprimir_automato(automato)
    print
    testar_automato(automato, aceita, rejeita)

def testar_er(params):
    """Função não implementada"""
    print "Função não implementada"

def converter_afn_para_afd(params):
    """Função não implementada"""
    print "Função não implementada"

def converter_afd_para_er(params):
    """Função não implementada"""
    print "Função não implementada"

def converter_er_para_afn(params):
    """Função não implementada"""
    print "Função não implementada"

def minimizar_afd(params):
    """Função não implementada"""
    print "Função não implementada"

def testar_equivalencia_afd(params):
    """Função não implementada"""
    print "Função não implementada"


# Interpretador de comandos

class InterpretadorComandos:
    def __init__(self):
        self.comandos = {}

    def add_comando(self, comando, func):
        self.comandos[comando] = func

    def ajuda(self, comandos):
        if len(comandos) == 0:
            self.exibir_comandos()
        else:
            if self.comandos.has_key(comandos[0]):
                print '* Ajuda do comando', comandos[0], '*'
                doc = self.comandos[comandos[0]].__doc__
                print doc if doc != None else "Sem descrição"
                
            else:
                print 'Comando inválido:', comandos[0]
                self.exibir_comandos()

    def exibir_comandos(self):
        print 'Para mais informações, digite ajuda nome-do-comando'
        print 'Comandos disponíveis:'
        for comando in sorted(self.comandos.keys()):
            print ' ', comando

    def executar(self, parametros):
        if len(parametros) == 0 or parametros[0] == 'ajuda':
            self.ajuda(parametros[1:])
        else:
            if self.comandos.has_key(parametros[0]):
                self.comandos[parametros[0]](parametros[1:])
            else:
                self.ajuda(parametros)
        
# Função principal
def main():
    import sys

    inte = InterpretadorComandos()
    inte.add_comando('testar-afd', testar_afd)
    inte.add_comando('testar-afn', testar_afn)
    inte.add_comando('testar-er', testar_er)
    inte.add_comando('converter-afn-para-afd', converter_afn_para_afd)
    inte.add_comando('converter-afd-para-er', converter_afd_para_er)
    inte.add_comando('converter-er-para-afn', converter_er_para_afn)
    inte.add_comando('minimizar-afd', minimizar_afd)
    inte.add_comando('testar-equivalencia-afd', testar_equivalencia_afd)

    inte.executar(sys.argv[1:])

if __name__ == '__main__':
    main()
